(in-package :cm)

;;; lock.lisp
;;; a time-based locking facility.  A locked datum becomes automatically
;;; unlocked after a specified period of time. algorithms can use this
;;; facility to "claim" values that should not be used for a specified
;;; unit of time:
;;;
;;; (locked? <thing> <time>)
;;; Returns true if <thing> is locked at <time>.
;;;
;;; (lock <thing> <start-time> <end-time>)
;;; Claims <thing> between times <start-time> and <end-time>
;;;
;;; (unlock <thing> <time>)
;;; Explicitly unlocks <thing> at <time>.  This is normally not necessary
;;; since locking expires automatically as time is incremented.
;;;
;;; (unlock-all)
;;; Unlocks everything. An algorithm should use this
;;; to reinitialize the locking mechanism before running.
;;;

(defclass lock ()
  ((table :initarg table :initarg :table :accessor lock-table)
   (count :initform 0 :initarg count :initarg :count :accessor lock-count)
   (maximum :initarg maximum :initarg :maximum :accessor lock-maximum)))

(defmethod initialize-instance :after ((object lock) &rest args)
  (declare (ignore args))
  (unless (slot-boundp object 'table)
    (setf (slot-value object 'table) (make-hash-table)))
  (unless (slot-boundp object 'maximum)
    (setf (slot-value object 'maximum) most-positive-fixnum)))

(defun make-lock (&rest args) (apply #'make-instance 'lock args))

(defparameter *lock* (make-lock))

(defun locked? (thing time &optional (lock *lock*)
                      &aux (table (lock-table lock)))
  (and (setf thing (gethash thing table))
       (<= (car thing) time (cdr thing))))

(defun lock (thing from to &optional (lock *lock*)
                   &aux (table (lock-table lock)))
  ;(format t "~%Locking ~A from ~S to ~S" thing from to ) 
  (let ((v (gethash thing table)))
    (unless v 
      (setf (gethash thing table) (setf v (cons nil nil)))
      (incf (lock-count lock)))
    (setf (car v) from (cdr v) to)
    thing))

(defun unlock (thing time &optional (lock *lock*)
                     &aux (table (lock-table lock)))
  (let ((v (gethash thing table)))
    (when v (setf (car v) time (cdr v) -1)
          (decf (lock-count lock)))
    thing))

(defun unlock-all (&optional (lock *lock*)) 
  (clrhash (lock-table lock))
  (setf (lock-count lock) 0))

(defun unlocked-item (stream time &optional (lock *lock*))
  (loop for thing = (item stream)
        unless (locked? thing time lock)
        return thing))

#|

;; locker1 displays the locking mechanism at work: Run locker1 0

(algorithm locker1 midi-note (length 100 initializer 'unlock-all amplitude .5)
  (loop do
        (setf note (item (notes c4 d ef f g af bf c5 in random)))
        while (locked? note time)
        do
        (format t "~%Skipping locked note ~S at time ~S." note time))
  (setf duration (item (rhythms s e q in random)))
  (setf rhythm (item (rhythms s e q in random)))
  (format t "~%Locking ~S from ~S to ~S." note time (+ time duration))
  (lock note time (+ time duration)))


;; locker2 is the same example without print out: mix locker2 0

(algorithm locker2 midi-note (length 100 initializer 'unlock-all amplitude .5)
  (loop do
        (setf note (item (notes c4 d ef f g af bf c5 in random)))
        while (locked? note time))
  (setf duration (item (rhythms s e q in random)))
  (setf rhythm (item (rhythms s e q in random)))
  (lock note time (+ time duration)))


|#
