;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :cm)


;;; advise midi menu to use stella streams.
#+mcl
(eval-when (load eval)
  (set-open/close-menu-item-function
   #'(lambda () 
       (set-output-stream
        (open-event-stream (find-syntax 'midi))))
   #'(lambda () (close-event-stream (find-syntax 'midi) ':force)))
  )

;;;
;;; This implementation distinguished between MIDI managers that
;;; support future messages scheduling (Apple) and bozo managers
;;; that do not (NeXTStep, Windows, SGI ...). Where scheduling is supported
;;; write-event for midi-port sends a NoteOn/NoteOff pair immediately.
;;; Where scheduling is not supported, the NoteOn is sent and the
;;; future NoteOff is inserted into CM's scheduler. This noteOff is
;;; held in resources to avoid consing.  Regardless of the MIDI
;;; manager's capabilities, the write-event method for midi-files
;;; must always schedule note offs.  It optimizes this by maintaining 
;;; a queue of pending messages  that it flushes whenever it writes
;;; a note.  For Bozo implementations, the same
;;; resource used by the midi-port is used.  The other implementation
;;; use the scheduler's cons cell resource.
;;;

#+(and apple (not midi-testing))
(progn
  ;; noteOff resource is just (time . message)
  (defvar *midi-resource* *scheduler-resource*)
  (defmacro resourced-midi-message (res) `(cdr ,res))
  (defmacro resourced-midi-time (res) `(car ,res))
  )

#-(and apple (not midi-testing))
(progn
  ;; noteOffs need a fullblown class because CM's scheduler is used.
  (defparameter *midi-resource* nil)
  (defclass resourced-midi (timed-object)
    ((datum :initarg :datum )))
  (defmacro resourced-midi-message (res) `(slot-value ,res 'datum))
  (defmacro resourced-midi-time (res) `(slot-value ,res 'time))
  (defresource *midi-resource* nil :size 64
    :constructor #'(lambda (&rest args) 
                     args (make-instance 'resourced-midi)))
  (eval-when (load eval)
    (initialize-resource *midi-resource* 64))
  )

(defmacro %scaled-time (time stream)
  `(round
    (if (slot-boundp ,stream 'scaler)
      (* (the number (slot-value ,stream 'scaler) )
         (the number ,time))
      (the number ,time))))

(defmacro flush-pending-offs (stream &optional time
                                     &aux (svar (gensym))
                                     (fvar (gensym)))
  `(let* ((,svar ,stream)
          (,fvar (slot-value ,svar 'stream)))
     (loop with last = (slot-value ,svar 'last-output-time)
           and queue = (slot-value ,svar 'releases)
           and scaler = (slot-value ,svar 'scaler)
           for off = (car (bag-elements queue))
           while ,(if time
                    `(and off (<= (resourced-midi-time off) ,time))
                    'off)
           do
           (bag-dequeue queue)
           ;(write-trace (resourced-midi-time off)
           ;             (round (* (- (resourced-midi-time off) last)
           ;                       scaler)))
           (%midifile-write-message
            (resourced-midi-message off) 
            ,fvar
            (round (* (- (resourced-midi-time off) last)
                      scaler)))
           (setf last (resourced-midi-time off))
           (setf (resourced-midi-message off) nil)
           (fast-deallocate-resource *midi-resource* nil off)
           finally (return last))))

(defsyntax midi 
  :pathname (pathname *default-midi-pathname*)
  :stream-types `((midi-port)
                  (midi-file
                   ,(pathname-type *default-midi-pathname*))))

(defmethod find-stream ((object midi) &optional arg)
  (find-stream (find-class 'midi-port arg)))

(defmethod open-event-stream ((object midi) 
                              &optional (mode ':default) &rest args)
  (declare (ignore mode))
  (apply #'open-event-stream (find-class 'midi-port) ':open args))

(defmethod close-event-stream ((object midi) &optional mode)
  (let ((stream (find-stream (find-class 'midi-port))))
    (when stream (close-event-stream stream mode))))

(defclass midi-stream (event-stream)
  ((syntax :initform (find-syntax ':midi))
   (scaler :accessor midi-stream-scaler)
   (chanmap :accessor midi-stream-chanmap
            ; :initform *default-midi-channel-map*
            :initarg chanmap :initarg :chanmap)
   (releases :initform (make-bag) :accessor midi-stream-releases)))

(defobject midi-port (midi-stream)
  ((direction :initform ':io)
   (to :initarg :port :initarg port
       :initarg :to :initarg to
       :accessor midi-port-to))
  (:parameters start end timescale to chanmap scheduling))

(defmethod default-stream-name ((object midi-port))
  (let ((count (count-streams (class-of object))))
    (format nil "Midi~@[-~D~]" (and (> count 0) (1+ count)))))

(defmethod single-stream? ((object midi-port)) 
  t)

(defmethod print-object ((object midi-port) stream)
  (format stream "#<Port: ~A to ~A~@[ (shut)~]>"
          (object-name object)
          (slot-value-or-default object 'to +slot-unset+)
          (not (slot-boundp object 'stream))))

(defmethod open-event-stream ((stream midi-port) &optional
                              (mode ':default) &rest args)
  ;; open when mode is :default or :open
  (unless (eq mode ':initialize)
    (let ((port (getf args 'port (slot-value-or-default stream 'to))))
      (setf port (midi-open :port port))
      (setf (slot-value stream 'to) port)
      (setf (slot-value stream 'stream) port))
    stream))

(defmethod initialize-stream-for-processing ((object midi-port))
  (let ((map (slot-value-or-default object 'chanmap nil))
        (tscale (slot-value-or-default object 'timescale 1)))
    (when (consp map)
      (setf (slot-value object 'chanmap) (apply #'vector map)))
    (if (eq *clock-mode* ':seconds)
      (setf (slot-value object 'scaler) (* 1000 tscale))
      (if (/= tscale 1)
        (setf (slot-value object 'scaler) tscale)
        (slot-makunbound object 'scaler)))
    (forget-bag (slot-value object 'releases))
    (midi-set-time 0)
    t))

(defmethod close-event-stream ((stream midi-port) &optional mode)
  (when (eq mode ':force)
    (slot-makunbound stream 'stream)
    (and (midi-close) t)))

(defobject midi-file (midi-stream event-file)
  ((last-output-time :initform 0.0)
   (divisions :initform 96 :initarg divisions
   	                  :initarg :divisions)
   (timesig :initform '(4 4) :initarg timesig :initarg :timesig)
   (tempo :initform 120 :initarg tempo :initarg :tempo)
   (element-type :initform '(unsigned-byte 8)))
  (:parameters start end timescale timesig tempo divisions chanmap))

(defmethod initialize-stream-for-processing ((object midi-file))
  (let ((f (slot-value object 'stream))
        (d (slot-value object 'divisions))
        (v (/ 60 (slot-value object 'tempo)))
        m)
    (let ((map (slot-value-or-default object 'chanmap nil)))
      (when (consp map)
        (setf (slot-value object 'chanmap) (apply #'vector map))))
    (forget-bag (slot-value object 'releases))
    (setf (slot-value object 'last-output-time) 0)
    ;(setf (slot-value object 'scaler)  ; .5=mm120
    ;      (* (/ d .5) (slot-value-or-default object 'timescale 1)))
    (setf (slot-value object 'scaler)  ; .5=mm120
          (* (/ (* d 2)
                (if (eq *clock-mode* ':seconds) 1 1000))
             (slot-value-or-default object 'timescale 1)))
    (%midifile-write-header f 0 1 d)
    (%midifile-write-track-header f 0)
    ;(multiple-value-setq (m d) (make-tempo-change 500000))
    (multiple-value-setq (m d) 
      (make-tempo-change  (floor (* 1000000 v))))
    (%midifile-write-message m f 0 d)
    ;(multiple-value-setq (m d) (make-time-signature 4 4))
    (multiple-value-setq (m d)
      (apply #'make-time-signature (slot-value object 'timesig)))
    (%midifile-write-message m f 0 d)
    t))

(defmethod deinitialize-stream-for-processing ((stream midi-file))
  (let ((f (slot-value stream 'stream)))
    (flush-pending-offs stream)
    (%midifile-write-message (make-eot) f 0 nil)
    (file-position f +midifile-header-length+)
    (%midifile-write-track-header
     f (- (file-length f)
          (+ +midifile-header-length+ +miditrack-header-length+)))
    #+mcl (ccl:set-mac-file-type f "Midi")
    t))

;;;
;;; midi listening/writing
;;;

#+(or mcl aclpc (and next excl) (and |NeXT| akcl) (and next clisp))
(defmethod play-using-syntax ((syntax midi) file &rest args)
  (unless (midi-open?) (midi-open :port (ask-port)))
  (apply #'midifile-play file
         (canonicalize-pairs args nil nil :keyword)))

#+sgi
(defmethod play-using-syntax ((syntax midi) file &rest pairs)
  (let* ((bg? (pair-value 'background pairs *post-processing-mode*))
         (pfile (namestring
                 (merge-pathnames "playmidifile" *cm-site-directory*)))
         (command (format nil "~A ~A~@[ &~]"
                          pfile (namestring (truename file)) bg?))) 
    (tell-user "Playing: ~A~%" command)
    (shell command)
    file))
                   
#-(or sgi mcl aclpc (and next excl) (and |NeXT| akcl) (and next clisp))
(defmethod play-using-syntax ((syntax midi) file &rest args)
  (declare (ignore file args))
  (format t "~%Don't know how to play midi files in this port.~%"))

(defmethod import-using-syntax ((syntax midi) file &rest args)
  (apply #'midifile-import file 
         (canonicalize-pairs args nil nil :keyword)))

;;;
;;; midi event classes
;;;

(defobject midi-message (rhythmic-element)
  ((message :initform 0 :initarg :message :initarg message
            :accessor midi-message)
   (data :initform nil :initarg data :initarg :data))
  (:parameters rhythm message))

(defobject midi-note (note)
  ((note :initarg note :initarg :note :accessor midi-note)
   (amplitude :initarg amplitude :initarg :amplitude
              :accessor midi-amplitude :initform 64)
   (duration :initarg :duration :initarg duration 
             :accessor midi-duration)
   (channel :initform 0 :initarg :channel :initarg channel
            :accessor midi-channel)
   (release :initarg :release :initarg release :accessor midi-release
            :initform 127))
  (:parameters note rhythm duration amplitude channel release))

(defmethod print-object ((object midi-message) stream)
  (midi-print-message (slot-value object 'message) nil :stream stream 
                      :message-data (slot-value object 'data)))
 
(defmethod print-object ((object midi-note) stream)
  (let* ((n (slot-value-or-default object 'note "---"))
         (r (slot-value-or-default object 'rhythm "------"))
         (d (if (slot-boundp object 'duration)
              (slot-value object 'duration) r))
         (a (slot-value-or-default object 'amplitude 64))
         (c (slot-value-or-default object 'channel 0)))
  (format stream 
          "#<MIDI-NOTE |~3:@A|~:[~6:@A~;~6,3F~]|~
           ~:[~6:@A~;~6,3F~]|~:[~3:@A~;~6,3F~]|~2:@A|>"
          (if (floatp n) (scale-degree n *standard-scale*) n)
          (floatp r) r (floatp d) d (floatp a) a c)))

(defmethod real-time-loop ((object container) (stream midi-port))
  (let* ((*merge* object)
         (okay nil)
         ;(time (slot-value object 'time))
         (base (midi-get-time))
         (this (+ base (slot-value (peek-object object) 'time)))
         (end (slot-value-or-default stream 'end))
         event)
    #+cltl2
    (declare (dynamic-extent *merge*))
    (unwind-protect 
      (progn
        (when end (incf end this))
        (loop
          #+mcl
          (process-wait "Waiting"
                        #'(lambda (m) (> (midi-get-time) m))
                        this)
          ;(loop when (>= (midi-get-time) this)
          ;      do (progn (format t "~%~a ~a" (midi-get-time) this)
          ;                (return)))
          ;; can't use process-wait in acl because the scheduluer quits
          ;; if all processes are waiting.  mp:process-allow-schedule is
          ;; used to allow other processes to run. get rid of this if it
          ;; introduces too much latency.
          #+excl
          (loop until (> (midi-get-time) this)
                do (mp:process-allow-schedule))
          (cond ((setf event (score-select object this))
                 (write-event event stream)
                 ;(setf this (+ base (slot-value (peek-object object) 
                 ;                               'time)))
                 (incf this (slot-value event 'rhythm))
                 (score-unset event))
                ((and end (< this end))
                 (setf this (midi-get-time)))
                (t (return))))
        (setf okay t))
      (if okay
        (progn
          (score-unset object nil)
          (close-event-stream stream nil)
          (post-process-stream stream nil))
        (progn
          (warn-user "Processing of ~S unexpectedly terminated.~&~
                      Performing recursive unset..." object)
          (deallocate-all-objects *midi-resource*)
          (score-unset object t)
          (close-event-stream stream ':error))))))

(defmethod setable-slots ((object midi-message))
  '(rhythm message data))

(defmethod setable-slots ((object midi-note))
  '(rhythm note duration amplitude channel release))

(defun write-trace (object time)
  (format t "~%@~:[~7:@A~;~7,3F~]	~S" (floatp time) time object))

#+(and apple (not midi-testing))
(defmethod write-event ((object midi-note) (stream midi-port))
  (let ((keynum (or (scale-degree (slot-value object 'note) 
                              *chromatic-scale*)
                    (error "No MIDI key number for note value: ~S"
                            (slot-value object 'note)))))
    (declare (fixnum keynum) (optimize (speed 3) (safety 0)))
    (if (<= 0 keynum 127)
      (let* ((time1 (slot-value object 'time))
             (time2 (+ time1 (if (slot-boundp object 'duration)
                               (slot-value object 'duration)
                               (slot-value object 'rhythm))) )
             (scaler? (slot-value-or-default stream 'scaler))
             (channel 
              (svref (the (vector t 32)
                       (slot-value-or-default 
                        stream 'chanmap *default-midi-channel-map*))
                     (slot-value object 'channel)))
             (amplitude (slot-value object 'amplitude))
             (release (slot-value object 'release))
             (noteon (%midi-note 9 channel keynum
                                 (if (< amplitude 1) 
                                   (floor (* amplitude 127))
                                   (if (integerp amplitude)
                                       (logand amplitude 127)
                                       (error "No MIDI velocity for amplitude ~
                                              value: ~S" amplitude)))))
             (noteoff (%midi-note 8 channel keynum
                                  (if (< release 1) 
                                    (floor (* release 127))
                                    (logand release 127))))
             flag)
        (when scaler?
          (setf time1 (* scaler? time1) time2 (* scaler? time2)))
        ;(write-trace noteon (round time2))
        (setf flag (%midi-write-message noteon (round time1)))
        ;(write-trace noteoff (round time2))
        (when (= flag +midi-success+)
          (setf flag (%midi-write-message noteoff (round time2))))
        (unless (= flag +midi-success+)
          (error "%midi-write-message failed with ~D" flag)))
      (unless (= keynum most-negative-fixnum)
         (error "No MIDI key number for note value: ~S"
                (slot-value object 'note))))))

#-(and apple (not midi-testing))
(progn
  
  (defmethod write-event ((object resourced-midi) (stream midi-port))
    (let ((flag (%midi-write-message 
                 (resourced-midi-message object)
                 (%scaled-time (resourced-midi-time object) stream))))
      (unless (= +midi-success+ flag)
        (error "%midi-write-message failed with ~D")))
    (fast-deallocate-resource *midi-resource* NIL object))
  
  (defmethod write-event ((object midi-note) (stream midi-port))
    (let ((time (slot-value object 'time))
          (queue (container-bag *merge*))
          (keynum (or (scale-degree (slot-value object 'note) 
                              *chromatic-scale*)
                      (error "No MIDI key number for note value: ~S"
                             (slot-value object 'note)))))
      (declare (fixnum keynum) ; (optimize (speed 3) (safety 0))
               )
      (if (<= 0 keynum 127)
        (let* ((c (svref (the (vector t 32)
                           (slot-value-or-default 
                            stream 'chanmap *default-midi-channel-map*))
                         (slot-value object 'channel)))
               (a (slot-value object 'amplitude))
               (r (slot-value object 'release))
               (f (%midi-write-message 
                   (%midi-note 9 c keynum 
                               (if (< a 1) 
                                   (floor (* a 127))
                                   (if (integerp a) 
                                       (logand a 127)
                                       (error "No MIDI velocity for amplitude ~
                                               value: ~S" a))))
                   (%scaled-time time stream))))
          (declare (fixnum c f))
          (if (= f +midi-success+)
            (let ((off (fast-allocate-resource *midi-resource*)))
              (setf (resourced-midi-time off)
                    (+ time 
                       (the number
                         (if (slot-boundp object 'duration)
                           (slot-value object 'duration)
                           (slot-value object 'rhythm))))
                    (resourced-midi-message off)
                    (%midi-note 8 c keynum
                                (if (< r 1) (floor (* r 127))
                                    (logand r 127))))
              (bag-enqueue off queue resourced-midi-time t))
            (error "%midi-write-message failed with error: ~D" f)))
        (unless (= keynum most-negative-fixnum)
          (error "No MIDI key number for note value: ~S"
                 (slot-value object 'note))))))
  )

(defmethod write-event ((object midi-note) (file midi-file))
  (let ((time (slot-value object 'time))
        (offs (slot-value file 'releases))
        (scaler (slot-value file 'scaler))
        (stream (slot-value file 'stream))
        (keynum (scale-degree (slot-value object 'note)
                              *chromatic-scale*))
        last)
    (declare (fixnum keynum) (number time last scaler)
             (optimize (speed 3) (safety 0)))
    (setf last
          (if (bag-elements offs)
            (flush-pending-offs file time)
            (slot-value file 'last-output-time)))
    (if (< keynum 0) ; rest
      (setf (slot-value file 'last-output-time) time)
      (let* ((c (svref (the (vector t 32)
                         (slot-value-or-default 
                          file 'chanmap *default-midi-channel-map*))
                       (slot-value object 'channel)))
             (a (slot-value object 'amplitude))
             (r (slot-value object 'release))
             (v (if (< a 1) (floor (* a 127)) a)))				 
        (declare (fixnum c v r))
        (if (> time last)
          (%midifile-write-message 
           (%midi-note 9 c keynum v) stream
           (round
            (the number (* (the number (- time last)) scaler))))
          (%midifile-write-message (%midi-note 9 c keynum v) stream))
        (setf (slot-value file 'last-output-time) time)
        (let ((off (fast-allocate-resource *midi-resource*)))
          (setf (resourced-midi-time off)
                (+ time 
                   (the number
                     (if (slot-boundp object 'duration)
                       (slot-value object 'duration)
                       (slot-value object 'rhythm))))
                (resourced-midi-message off) 
                (%midi-note 8 c keynum (if (< r 1) (floor (* r 127)) 
                                      (logand r 127))))
          (bag-enqueue off offs resourced-midi-time))))))

(defmethod write-event ((object midi-message) (port midi-port))
  (declare (optimize (speed 3) (safety 0)))
  (let ((f (%midi-write-message
            (slot-value object 'message) 
            (%scaled-time (slot-value object 'time) port))))
    (unless (= f +midi-success+)
      (error "%midi-write-message failed with error: ~s" f))
    (let ((data (slot-value object 'data)))
      (when data
        (let ((q (%scaled-time (slot-value object 'time) port)))
          (dolist (d data) (%midi-write-message d q))))
      object)))

(defmethod write-event ((object midi-message) (file midi-file))
  
  (let ((time (slot-value object 'time))
        (last (slot-value file 'last-output-time)))
    (declare (number time last) (optimize (speed 3) (safety 0)))
    (cond ((> time last)
           (%midifile-write-message 
            (slot-value object 'message)  
            (slot-value file 'stream)
            (round (the number (* (the number (- time last) )
                                  (the number (slot-value file 'scaler))))))
           (setf (slot-value file 'last-output-time) time))
          (t
           (%midifile-write-message (slot-value object 'message)
                                    (slot-value file 'stream))))
    (let ((data (slot-value object 'data)))
      (when data
        (%midifile-write-message-data data (slot-value file 'stream))))
    object))

;;;
;;;
;;;

(defun midifile-import (midifile &key (track 0) (start 0.0) 
                                 end channels container
                                 (note-type ':note))
  (declare (ignore channels))
  (let ((list '())
        (hash (make-hash-table))
        (time 0.0)
        (notefn (ecase note-type
                  ((:note note) #'scale-note)
                  ((:pitch pitch) #'scale-pitch)
                  ((:degree degree) nil))))
    (with-open-midifile (file midifile :track track :ttype :seconds)
      (when (> start 0.0)
        (midifile-set-track-start file start))
      (flet
        ((fn (m r &aux (s (%message-status m)) n)
           (declare (float time r) (fixnum m s)
                    (optimize (speed 3) (safety 0)))
           (incf time r)
           (cond ((or (= s +midi-note-off+)
                      (and (= s +midi-note-on+) 
                           (= 0 (%message-velocity m))))
                  (let ((o (pop (gethash (%message-hash m) hash))))
                    (if o 
                      (setf (midi-duration o) (- time (object-time o)))
                      (warn "No NoteOn for channel ~D keynum ~D." 
                            (%message-channel m) 
                            (%message-keynum m)))))
                 ((= s +midi-note-on+)
                  (setf n
                        (make-instance 'midi-note
                          :start time
                          :note (if notefn
                                  (funcall notefn
                                           (%message-keynum m)
                                           *chromatic-scale*)
                                  (%message-keynum m))
                          :channel (%message-channel m)
                          :amplitude (/ (%message-velocity m) 127.0)
                          ))
                  (push n (gethash (%message-hash m) hash)))
                 (t 
                  (setf n (make-instance 'midi-message
                            :start time :message m))))
           (when n
             (setf (element-flags n) 0)
             (when list
               (setf (object-rhythm (car list)) 
                     (- time (object-time (car list)))))
             (push n list))))
        (midifile-map-track file #'fn))
      ;; last object parsed has no rhythm, set it to its duration or 0
      (setf (object-rhythm (car list))
            (if (slot-exists-p (car list) 'duration)
              (midi-duration (car list))
              0.0))
      ;; no events can begin past the current time but keep
      ;; mapping until all note offs have been found.
      (when end
        (let ((n 0))
          (maphash #'(lambda (k v) k (incf n (length v))) hash)
          (when (> n 0)
            (block drain
              (flet ((off (m r &aux (s (%message-status m)))
                       (incf time r)
                       (when (or (= s +midi-note-off+)
                                 (and (= s +midi-note-on+) 
                                      (= 0 (%message-velocity m))))
                         (let ((o (pop (gethash (%message-hash m) hash))))
                           (when o 
                             (setf (midi-duration o) 
                                   (- time (object-time o)))
                             (when (= (decf n) 0) (return-from drain nil)))))))
                (midifile-map-track file #'off)))
            (setf (object-rhythm (car list)) 
                  (midi-duration (car list)))
            (when (> n 0) (format t "Didn't find some note offs!")))))
      ;; find a new name for the conainer
      (unless container
        (let  ((root (format nil "FROM-~:@(~A~)" 
                             (pathname-name (midifile-pathname file)))))
          (loop with name = root and num = 0
                for sym = (find-symbol name)
                while (and sym (find-object sym))
                do (setf name (format nil "~A-~D" root (incf num)))

                finally
                (setf container
                      (make-instance 'thread :id (intern name)))
                )))
      (remove-all-objects container)
      (add-objects (nreverse list) container nil :copy-first nil)
      container)))



