;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: taube@uiuc.edu
;;; **********************************************************************

(in-package :common-music)

(defclass essential-scale ()
  ((lowest :accessor scale-lowest :initarg :lowest)
   (name :accessor scale-name :initarg :name)
   (divisions-per-octave :accessor scale-divisions-per-octave
                        :initarg :divisions-per-octave)
   (number-of-octaves :accessor scale-number-of-octaves
                      :initarg :number-of-octaves)
   (number-of-entries :accessor scale-number-of-entries
                      :initarg :number-of-entries)
   (entries :accessor scale-entries 
            :initform (make-hash-table :size 512))
   (tokens :accessor scale-tokens :initarg :tokens :initform nil)))

(defmethod print-object ((object essential-scale) stream)
  (format stream "#<~A ~:(~A~)>" (class-name (class-of object))
                                 (scale-name object)))
(defclass general-scale (essential-scale)
  ((pitches :accessor scale-pitches :initarg :pitches)))

(defclass octave-mixin ()
  ((octave-ratio :accessor scale-octave-ratio :initarg :octave-ratio)))
   
(defclass gapped-scale (essential-scale octave-mixin)
  ((gaps :accessor scale-gaps :initarg :gaps))) 

(defclass equal-tempered-scale (essential-scale octave-mixin)
  ((interval-ratio :accessor scale-interval-ratio 
                   :initarg :interval-ratio)))

(defvar *scales* (make-hash-table)
  "Table of musical scales.")

(defun list-all-scales ()
  (let ((scales ()))
    (maphash #'(lambda (k v) k (push v scales)) *scales*)
    (nreverse scales)))

(defun find-scale (name &optional (errorp t))
  (or (gethash name *scales*)
      (if errorp 
          (error "No scale named ~S." name)
        nil)))

(defun cents (cents)
  ;(expt 10 (* cents (/ (log 2 10) 1200)))
  (expt 2.0 (/ cents 1200)))

(defun centify (scaler)
  ;(* (/ 1200 (log 2 10)) (log scaler 10))
  (round (* (log scaler 2.0) 1200))
  )

(defconstant *rest-name* 'r)        ; these should be scale specific
(defconstant *rest-degree* most-negative-fixnum)
(defconstant *rest-pitch* most-negative-single-float)

(defun restp (x)
  (cond ((eq x *rest-name*) t)
        ((integerp x) (= x *rest-degree*))
        ((floatp x) (= x *rest-pitch*))
        (t nil)))

;;;
;;;
;;;

(defun make-scale (&key name class entries (number-of-octaves nil nop)
                        lowest-pitch octave-ratio interval-ratio
                        first-octave-number divisions-per-octave
                        suboctave-enharmonic 
                        superoctave-enharmonic)
  ;; normalize entries: ((&rest notes) &key pitch scaler ratio)
  (setf entries
       (loop for spec in entries
             collect
             (cond ((consp spec)
                    (if (consp (car spec)) 
                        spec
                      (if (find (cadr spec) '(:pitch :scaler :ratio))
                          (cons (list (car spec)) (cdr spec))
                        (list spec))))
                   (t (list (list spec))))))
  (let ((state 0) 
        octave-suffixes scale)
    (when (null nop) 
      (setf number-of-octaves 9))
    (if (null divisions-per-octave)
        (setf divisions-per-octave (length entries))
     (unless number-of-octaves
       (setf number-of-octaves 
             (/ (length entries) divisions-per-octave))))
    (if interval-ratio
        (if octave-ratio 
            (error "Found exclusive keywords :octave-ratio, ~
                    :interval-ratio.")
          (setf octave-ratio (expt interval-ratio divisions-per-octave)))
      (when octave-ratio
        (setf interval-ratio 
              (expt octave-ratio (/ 1 divisions-per-octave)))))
    (cond ((and (or (stringp first-octave-number) 
                    (symbolp first-octave-number))
                (string= (string first-octave-number) "00"))
           (setf octave-suffixes
             (cons '|00|
                (loop for i from 0 below (1- number-of-octaves)
                      collect i))))
          ((integerp first-octave-number)
           (setf octave-suffixes
             (loop repeat number-of-octaves
                   collect first-octave-number
                   do (incf first-octave-number))))
          ((null first-octave-number)
           (setf octave-suffixes
             (loop for i below number-of-octaves collect i)))
          (t (error "Bogus first-octave-number ~S."
                    first-octave-number)))
    ;; parse entries, determining a scaler and frequency value for each
    ;; entry. the state variable keeps track of which keywords are used
    ;; in the specs; more than one state is an error.
    (setf entries
      (loop with accumlated-scaler = 1.0 and entry
            for spec in entries
            for position from 0
            do (multiple-value-setq (entry lowest-pitch accumlated-scaler
                                     state)
                 (parse-entry spec position lowest-pitch interval-ratio
                              accumlated-scaler state))
            collect entry
            finally (unless (member state '(0 1 2 4))
                      (error "Can't mix :scaler, :pitch and :ratio."))))
    ;(pprint entries)
    (unless class
      (setf class
        (if (= state 0) 'equal-tempered-scale
          (if octave-ratio 'gapped-scale 'general-scale))))
    (setf scale
      (make-instance class :name (or name (gentemp class))
                     :lowest lowest-pitch 
                     :number-of-octaves number-of-octaves
                     :divisions-per-octave divisions-per-octave))
    (fill-scale scale entries number-of-octaves
                octave-ratio interval-ratio octave-suffixes 
                suboctave-enharmonic superoctave-enharmonic)
    (setf (gethash name *scales*) scale)
    scale))

(defun parse-entry (spec position lowest-pitch interval-ratio
                         accumlated-scaler state)
  ;; parse spec, check keyword consistency across all the specs.
  (let ((notes (pop spec))
        pitch scaler)
    (unless (or (null spec) (= (length spec) 2))
      (error "Malformed keyword list: ~S" spec))
    (cond ((eq (car spec) ':pitch)
           (setf pitch (cadr spec) state (logior state #b001))
           (when (= position 0) (setf lowest-pitch pitch)))
          ((eq (car spec) ':scaler)
           (setf scaler (cadr spec) state (logior state #b010))
           (setf pitch (* lowest-pitch scaler)))
          ((eq (car spec) ':ratio)
           (setf scaler accumlated-scaler state (logior state #b100)
                 accumlated-scaler (* accumlated-scaler (cadr spec))))
          (spec
           (error "~S not a legal keyword specification." spec))
          (t 
           (when (logtest state #b001)
             (error "Missing :pitch value for ~S" spec))
           (setf scaler (expt interval-ratio position))))
    (when (= position 0)
      (unless (= scaler 1.0)
        (error "Scaler for first degree not 1.0: ~S" spec)))
    (unless lowest-pitch (error "Missing :lowest-pitch value."))

    (values (list notes position
                  (or pitch (* lowest-pitch scaler))
                  scaler) 
            lowest-pitch accumlated-scaler state)))

(defmethod fill-scale ((scale essential-scale)  &optional
                        entries number-of-octaves
                        octave-ratio interval-ratio octave-suffixes 
                        suboctave-enharmonic superoctave-enharmonic)
  (declare (ignore octave-ratio interval-ratio
                   suboctave-enharmonic superoctave-enharmonic))
  ;; each entry is a list: (names position pitch scaler)
  ;; make a token table to hold both note names and octave suffixes 
  ;; for resolving note names without an octave number. because of the 
  ;; octave "number" 00,  which the lisp reader rightly confuses with 0, 
  ;; we must hash octave suffixes to actual octave numbers ...Sooo, the
  ;; octave suffix is either a number or a symbol. its hash values are 
  ;; the starting step position in the scale. we also reverse hash the
  ;; octave's step position to return the octave token for fast lookup of
  ;; a note's octave designator. the step postions are all offset by 1024
  ;; to avoid collision with numerical octave hash keys.
  (when (> number-of-octaves 0)
    (let ((tokens (make-hash-table 
                    :size (+ (* number-of-octaves 2)
                                (loop for e in entries 
                                      sum (length (first e)))))))
      ;; link note names to position in octave
      (loop for e in entries 
           do (loop for n in (first e) 
                    do (setf (gethash n tokens) (second e))))
      ;; reverse link octaves to octave numbers
      (loop for s in octave-suffixes   
            for o below number-of-octaves
            do  (setf (gethash s tokens) o)
                (setf (gethash (+ o 1024) tokens) s))
      (setf (slot-value scale 'tokens) tokens)))
  scale)

(defmethod fill-scale :after ((scale octave-mixin)  &optional
                              entries number-of-octaves
                              octave-ratio interval-ratio octave-suffixes 
                              suboctave-enharmonic superoctave-enharmonic)
  (declare (ignore interval-ratio))
  ;; each entry is a list: (names position pitch scaler)
  (let ((divisions (slot-value scale 'divisions-per-octave))
        (table (slot-value scale 'entries)))
    ;; now we are ready to fill the actual entry hashtable.   
    ;; iterate over the octaves creating the actual note names in the
    ;; scale out of the note descritions and the octave suffixes.
    (loop with last = (1- (* number-of-octaves divisions))
          for octave-number below number-of-octaves
          for octave-scaler = (expt octave-ratio octave-number)
          for octave-step = (* octave-number divisions)
          do (loop for entry in entries
                   for step  = (+ (second entry) octave-step)
                   for pitch = (float (* (third entry) octave-scaler))
                   for notes = (make-note-names (first entry) 
                                                octave-number 
                                                octave-suffixes
                                                superoctave-enharmonic
                                                suboctave-enharmonic
                                                (= step 0) (= step last))
                   do ;(print (list notes pitch))
                             (add-scale-entries table step pitch notes))
          ;; add rest values to the scale.  these should really be scale
          ;; specific but currently they aren't...
          finally (add-scale-entries table *rest-degree* *rest-pitch* 
                                   (list *rest-name*)))
    (setf (slot-value scale 'octave-ratio) octave-ratio)
    (setf (slot-value scale 'number-of-entries)
      (* number-of-octaves divisions))
    scale))

(defmethod fill-scale :after ((scale equal-tempered-scale)  &optional
                              entries number-of-octaves
                              octave-ratio interval-ratio octave-suffixes 
                              suboctave-enharmonic superoctave-enharmonic)
  (declare (ignore entries number-of-octaves octave-ratio octave-suffixes 
                   suboctave-enharmonic superoctave-enharmonic))
  (setf (slot-value scale 'interval-ratio) interval-ratio))


(defmethod fill-scale :after ((scale gapped-scale)   &optional
                              entries number-of-octaves
                              octave-ratio interval-ratio octave-suffixes 
                              suboctave-enharmonic superoctave-enharmonic)
  (declare (ignore number-of-octaves
                   octave-ratio interval-ratio octave-suffixes 
                   suboctave-enharmonic superoctave-enharmonic))
  (let ((gaps (make-array (slot-value scale 'divisions-per-octave))))
    (loop for i from 0 as e in entries 
          do (setf (svref gaps i) (fourth e)))
    (setf (slot-value scale 'gaps) gaps)))


(defmethod fill-scale :after ((scale general-scale)  &optional
                              entries number-of-octaves
                              octave-ratio interval-ratio octave-suffixes
                              suboctave-enharmonic superoctave-enharmonic)
  (declare (ignore interval-ratio octave-ratio suboctave-enharmonic 
                   number-of-octaves
                   superoctave-enharmonic))
  (let ((divisions (slot-value scale 'divisions-per-octave))
        length table pitches)
    (setf length (length entries))
    (setf table (slot-value scale 'entries))
    (setf pitches (make-array length :element-type 'float))
    (setf (slot-value scale 'number-of-entries) length)
    (loop with last = (1- length)
          for entry in entries 
          for step from 0
          for octave-number = (floor step divisions)
          for pitch = (float (or (third entry)
                                 (error "Missing :pitch information for ~
                                         general-scale degree ~S." entry)))
          for notes = (make-note-names (first entry) 
                                       octave-number octave-suffixes
                                       nil nil (= step 0) (= step last))
          do (add-scale-entries table step pitch notes)
             (setf (svref pitches step) pitch)
          finally (add-scale-entries table *rest-degree* *rest-pitch* 
                                     (list *rest-name*)))
    (setf (slot-value scale 'pitches) pitches)
    scale))
    
    
(defun add-scale-entries (table degree pitch names)
  ;; create note and degree hash entries in scale. hash value for degree
  ;; is the cons (pitch . names).  hash value for name is (pitch . degree)
  ;; floating point pitch is a not hashed, each class of scale
  ;; implements its own methods to map frequency to note or degree.
  (setf (gethash degree table) (cons pitch names))
  (let ((entry (cons pitch degree))) 
    (dolist (name names) (setf (gethash name table) entry)))
  (values))

(defun make-note-names (names octave-number octave-suffixes
                        superoctave-enharmonic suboctave-enharmonic 
                        first? last?)
  (cond ((and first? suboctave-enharmonic)
         (setf names (remove suboctave-enharmonic names)))
        ((and last? superoctave-enharmonic)
         (setf names (remove superoctave-enharmonic names))))
  (loop for name in names
        for suffix = (nth (if (eq name superoctave-enharmonic) 
                              (1+ octave-number)
                            (if (eq name suboctave-enharmonic)
                                (1- octave-number)
                              octave-number))
                          octave-suffixes)
        collect (intern (concatenate 'string
                                     (string name) 
		                     (if (numberp suffix)
			                 (prin1-to-string suffix)
			               (string suffix)))) ))

;;;
;;; defscale and the standard chromatic scale definition
;;;

(defmacro defscale (name args &body body)
  (let (variable form)
    (if (find ':name args)
        (setf variable name)
      (setf args (list* ':name `(quote ,name) args) variable nil))
    (setf form    
      `(make-scale :entries
                   (list
                     ,@ (loop for form in body
                              if (and (consp form)
                                      (member (second form) 
                                              '(:scaler :pitch :ratio)))
                              collect `(list (quote ,(car form)) 
                                             ,@(cdr form))
                              else collect `(quote ,form)))
                   ,@args))
    (if variable
        `(defparameter ,variable ,form)
      form)))

(defvar *chromatic-scale*)
(defparameter *standard-scale* nil)

(eval-when (load eval)
  (setf *chromatic-scale*
        (defscale chromatic-scale 
                  (:lowest-pitch 8.175798 
                                 :octave-ratio (cents 1200)
                                 :suboctave-enharmonic 'bs
                                 :superoctave-enharmonic 'cf 
                                 :number-of-octaves 11
                                 :first-octave-number '|00|)
          (C CN BS)
          (CS DF)
          (D DN)
          (DS EF)
          (E EN FF)
          (F FN ES)
          (FS GF)
          (G GN)
          (GS AF)
          (A AN)
          (AS BF)
          (B BN CF)))
  (setf *standard-scale* *chromatic-scale*))

;;;
;;; scale referencing. scale-pitch returns the floating point frequency
;;; of a scale degree reference. scale-note returns the symbolic name of
;;; a scale degree reference. scale-degree returns the ordinal position
;;; of a scale degree reference. octave-and-interval returns the octave
;;; and interval position in the octave of a scale degree reference.
;;;

(defmethod scale-pitch ((ref symbol) (scale essential-scale))
  (car (gethash ref (slot-value scale 'entries))))

(defmethod scale-pitch ((ref integer) (scale essential-scale))
  (car (gethash ref (slot-value scale 'entries))))

(defmethod scale-pitch ((ref float) (scale essential-scale))
  ref)

;;;
;;;
;;;

(defmethod scale-note ((ref symbol) (scale essential-scale))
  ref)
  
(defmethod scale-note ((ref integer) (scale essential-scale))
  (cadr (gethash ref (slot-value scale 'entries))))

(defmethod scale-note ((ref float) (scale equal-tempered-scale))
  (let ((degree (round (log (/ ref (slot-value scale 'lowest))
			    (slot-value scale 'interval-ratio)))))
    (cadr (gethash degree (slot-value scale 'entries)))))

(defmethod scale-note ((ref float) (scale gapped-scale))
  (multiple-value-bind (oct int) (octave-and-interval ref scale)
    (cadr (gethash (+ (* oct (slot-value scale 'divisions-per-octave))
                      int)
                   (slot-value scale 'entries)))))

(defmethod scale-note ((ref float) (scale general-scale))
  (cadr (gethash (scale-degree ref scale)
                 (slot-value scale 'entries))))
;;;
;;;
;;;

(defmethod scale-degree ((ref symbol) (scale essential-scale))
  (cdr (gethash ref (slot-value scale 'entries))))

(defmethod scale-degree ((ref integer) (scale essential-scale))
  ref)

(defmethod scale-degree ((ref float) (scale equal-tempered-scale))
  (values (round (log (/ ref (slot-value scale 'lowest))
		      (slot-value scale 'interval-ratio)))))

(defmethod scale-degree ((ref float) (scale gapped-scale))
  (multiple-value-bind (oct int) (octave-and-interval ref scale)
    (+ (* oct (slot-value scale 'divisions-per-octave) )
       int)))

(defmethod scale-degree ((ref float) (scale general-scale))
  (loop with pitches = (slot-value scale 'pitches)
        for i from 0 below (slot-value scale 'number-of-entries)
        while (<= (svref pitches i) ref) 
        finally (return (1- i))))

;;;
;;;
;;;

(defmethod octave-and-interval ((ref integer) (scale essential-scale))
  (floor ref (slot-value scale 'divisions-per-octave)))

(defmethod octave-and-interval ((ref symbol) (scale essential-scale))
  (let ((degree (cdr (gethash ref (slot-value scale 'entries)))))
    (when degree
      (floor degree (slot-value scale 'divisions-per-octave)))))

(defmethod octave-and-interval ((ref float) (scale equal-tempered-scale))
  (let ((degree (round (log (/ ref (slot-value scale 'lowest))
			    (slot-value scale 'interval-ratio)))))
    (floor degree (slot-value scale 'divisions-per-octave))))

(defmethod octave-and-interval ((ref float) (scale gapped-scale))
  (let* ((lowest (slot-value scale 'lowest))
         (ratio (slot-value scale 'octave-ratio))
         (octnum (floor (log (/ ref lowest) ratio)))
         (octmul (expt ratio octnum)))
    ;; this test is a house of cards, at least for acl/next. the 
    ;; calculation of (* gap lowest octmul) below MUST be computed in
    ;; exactly the manner and order that fill-scale created the scale 
    ;; entry. lowest*octmul is actually a constant for the loop, but we
    ;; set octfrq=lowest*octmul and then use octfrq*gap in the loop, the
    ;; value arrived at is sometimes very slightly off and causes an
    ;; interval to be miscounted.
    (values octnum 
           (loop with gaps = (slot-value scale 'gaps)
                 for i below (slot-value scale 'divisions-per-octave)
                 while (<= (* (svref gaps i) lowest octmul) ref) 
                 finally (return (1- i))))))

(defmethod octave-and-interval ((ref float) (scale general-scale))
  (let ((step (scale-degree ref scale)))
    (if (> (slot-value scale 'number-of-octaves) 0)
        (floor step (slot-value scale 'divisions-per-octave))
     (values nil step))))

;;;
;;; use this function to check a scale to make sure that it can
;;; map pitches to degrees and notes.
;;;

(defun test-scale (scale &optional only-errors (fudge 0.0))
  (loop for step from 0
        for note = (scale-note step scale)
        while note
        for pitch = (+ (scale-pitch note scale) fudge)
        for check1 = (scale-note pitch scale)
        for check2 = (scale-degree pitch scale)
        for check3 = (and (eq check1 note) (eq check2 step))
        when (or (not only-errors) (not check3))
        do (format t "~&~A	~A	~12,6F ~:[!~;=~] ~A ~A"
                   step note pitch check3 check1 check2)))

;;;
;;; user functions all use *standard-scale*.
;;;

(defun in-scale (scale)
  (setf *standard-scale* (find-scale scale)))

(defmethod note ((note t) &optional (scale *standard-scale*))
  (scale-note note scale))


(defmethod pitch ((note t) &optional (scale *standard-scale*))
  (scale-pitch note scale))


(defmethod degree ((note t) &optional (scale *standard-scale*))
  (scale-degree note scale))


;;;
;;; predicates for comparing scale references. only need to use 
;;; in a scale if types are different or note names are used.
;;; methods are binary, the >, < and = forms are n-ary.
;;;

(defun scale<= (&rest args)
  (loop for (x y) on args by #'cdr while y 
        always (scale-leq x y)) )

(defmethod scale-leq ((x t) (y t) &optional (scale *standard-scale*))
  (<= (scale-degree x scale) (scale-degree y scale)))

(defmethod scale-leq ((x integer) (y integer) &optional scale)
   (declare (ignore scale))
  (<= x y))

(defmethod scale-leq ((x float) (y float) &optional scale)
   (declare (ignore scale))
  (<= x y))

(defun scale>= (&rest args) 
  (loop for (x y) on args by #'cdr while y 
        always (scale-geq x y)) )

(defmethod scale-geq ((x t) (y t) &optional (scale *standard-scale*))
  (>= (scale-degree x scale) (scale-degree y scale)))

(defmethod scale-geq ((x integer) (y integer) &optional scale)
   (declare (ignore scale))
  (>= x y))

(defmethod scale-geq ((x float) (y float) &optional scale)
   (declare (ignore scale))
  (>= x y))

(defun scale/= (&rest args) 
  (loop for (x y) on args by #'cdr while y 
        always (scale-neq x y)) )

(defmethod scale-neq ((x t) (y t) &optional (scale *standard-scale*))
  (/= (scale-degree x scale) (scale-degree y scale)))

(defmethod scale-neq ((x integer) (y integer) &optional scale)
   (declare (ignore scale))
  (/= x y))

(defmethod scale-neq ((x float) (y float) &optional scale)
   (declare (ignore scale))
  (/= x y))

(defun scale> (&rest args) 
  (loop for (x y) on args by #'cdr while y 
        always (scale-greaterp x y)))

(defmethod scale-greaterp ((x t) (y t) &optional (scale *standard-scale*))
  (> (scale-degree x scale) (scale-degree y scale)))

(defmethod scale-greaterp ((x integer) (y integer) &optional scale)
   (declare (ignore scale))
  (> x y))

(defmethod scale-greaterp ((x float) (y float) &optional scale)
   (declare (ignore scale))
  (> x y))

(defun scale< (&rest args)
  (loop for (x y) on args by #'cdr while y 
        always (scale-lessp x y)) )

(defmethod scale-lessp ((x t) (y t) &optional (scale *standard-scale*))
  (< (scale-degree x scale) (scale-degree y scale)))

(defmethod scale-lessp ((x integer) (y integer) &optional scale)
   (declare (ignore scale))
  (< x y))

(defmethod scale-lessp ((x float) (y float) &optional scale)
   (declare (ignore scale))
  (< x y))

(defun scale= (&rest args) 
  (loop for (x y) on args by #'cdr while y 
        always (scale-equalp x y)) )

(defmethod scale-equalp ((x t) (y t) &optional (scale *standard-scale*))
  (= (scale-degree x scale) (scale-degree y scale)))

(defmethod scale-equalp ((x integer) (y integer) &optional scale)
   (declare (ignore scale))
  (= x y))

(defmethod scale-equalp ((x float) (y float) &optional scale)
   (declare (ignore scale))
  (= x y))

(defun scale- (x y &optional (scale *standard-scale*))
  (- (scale-degree x scale) (scale-degree y scale)))

(defun scale-min (&rest notes)
  (reduce #'scale-lower notes))

(defmethod scale-lower (x y &optional (scale *standard-scale*))
  (if (< (scale-degree x scale) (scale-degree y scale)) 
    x y))

(defun scale-max (&rest notes)
  (reduce #'scale-higher notes))

(defmethod scale-higher (x y &optional (scale *standard-scale*))
  (if (> (scale-degree x scale) (scale-degree y scale)) 
    x y))

(defun scale-member (ref refs &optional (scale *standard-scale*))
  (find ref refs :test #'(lambda (x y) (scale-equalp x y scale))))


;;;
;;;
;;;

(defmethod transpose ((degree integer) (interval integer)
                      &optional scale)
  (declare (optimize (speed 3) (safety 0))
	    (ignore scale))
  (if (= degree *rest-degree*) *rest-degree*
    (+ (the integer degree)
       (the integer interval))))

(defmethod transpose ((name symbol) (interval integer) 
		      &optional (scale *standard-scale*))
  (declare (optimize (speed 3) (safety 0)))
  (if (eq name *rest-name*) *rest-name*
    (let* ((table (slot-value scale 'entries))
           (index (cdr (gethash name table)))
           (entry1 (gethash index table))
           (entry2 (gethash (+ index interval) table)))
      (or (nth (position name (cdr entry1)) (cdr entry2) )
          (first (cdr entry2))))))

;(defmethod transpose ((name symbol) (interval integer) 
;		      &optional (scale *standard-scale*))
;  (declare (optimize (speed 3) (safety 0)))
;  (scale-note (+ (the integer (scale-degree name scale))
;		 (the integer interval))
;	      scale))

(defmethod transpose ((pitch float) (interval integer) 
		      &optional (scale *standard-scale*))
  (declare (optimize (speed 3) (safety 0)))
  (if (= pitch *rest-pitch*) *rest-pitch*
    (scale-pitch (+ (the integer (scale-degree pitch scale)) 
                    (the integer interval))
                 scale)))

(defmethod transpose ((notes cons) (interval integer)
                      &optional (scale *standard-scale*))
  (mapcar #'(lambda (x) (transpose x interval scale)) notes))

;;;
;;; invert a scale reference around some point.
;;;
  
(defmethod invert ((note symbol) point 
                   &optional (scale *standard-scale*))
  (let ((off (scale-degree point scale)))
    (scale-note (+ off (* (- (scale-degree note scale) off) -1))
                scale)))

(defmethod invert ((note integer) point 
                   &optional (scale *standard-scale*))
  (let ((off (scale-degree point scale)))
    (scale-degree (+ off (* (- (scale-degree note scale) off) -1))
                  scale)))

(defmethod invert ((note float) point 
                   &optional (scale *standard-scale*))
  (let ((off (scale-degree point scale)))
    (scale-pitch (+ off (* (- (scale-degree note scale) off) -1))
                 scale)))

(defmethod invert ((notes cons) point
                   &optional (scale *standard-scale*))
  (if point
    (let ((degree (scale-degree point scale)))
      (mapcar #'(lambda (x) (invert x degree scale)) notes))
    (let* ((first (scale-degree (first notes) scale))
           (lowest (scale-degree (apply #'scale-lowest notes) scale))
           (delta (- (- (scale-degree (apply #'scale-highest notes) scale)
                        (- first lowest))
                     first)))
      (mapcar #'(lambda (x) 
                  (transpose (invert x first scale) delta scale))
              notes))))

;;;
;;; modes are transposeable subsets of scales.
;;;

(defmacro defmode (name &rest steps)
  (let ((args '()))
    (when (consp name)
      (setf args (loop for (k v) on (rest name) by #'cddr
                       collect `',k collect v)
            name (first name)))
    `(make-instance 'mode :name ',name :steps '(,@steps) ,@args)))

(defclass mode ()
  ((name :initarg name :initarg :name)
   (scale :initarg :scale :initarg scale :initform *standard-scale*)
   (on :initform nil :initarg :on :initarg on)
   (fit :initarg :steps :initarg steps)
   int octave steps))

(defmethod initialize-instance :after ((object mode) &rest args)
  (declare (ignore args))
  (let* ((steps (slot-value object 'fit))
         octave)
    (when (find-if #'symbolp steps)
      (let ((*standard-octave* 4)
            (scale (slot-value object 'scale))
            first note)
         (declare (special *standard-octave*))
         (setf first (parse-note (pop steps) scale))
         (setf steps
               (loop with last = first
                     for next in steps
                     do (setf note (parse-note next scale))
                     collect (scale- note last)
                     do (setf last note)))
         (setf (slot-value object 'on) first)))
    (setf octave (apply #'+ steps))
    (setf (slot-value object 'octave) octave)
    (setf (slot-value object 'steps) (length steps))
    (setf (slot-value object 'fit)
          (loop with s = steps and l = '() and i = 0
                while (< i octave) do 
                (loop for k from 0 by -1 repeat (pop s)
                      do (push k l) (incf i))
                finally (return (apply #'vector (nreverse l)))))
    (setf (slot-value object 'int)
          (apply #'vector (loop with i = 0
                                for s in steps
                                collect i do (incf i s))))
    (when (slot-value object 'on)
      (transpose object (slot-value object 'on)))
    (setf (gethash (slot-value object 'name) *scales*) object)))

(defmethod print-object ((object mode) stream)
  (let ((name (slot-value object 'name))
        (on (slot-value object 'on)))
    (if on
      (let ((n (string (scale-note on (slot-value object 'scale)))))
        (format stream "#<MODE: ~@(~A~) (on ~A)>" 
                name 
                (if (= (slot-value object 'octave)
                       (cm::scale-divisions-per-octave 
                        (slot-value object 'scale))) 
                  (subseq n 0 (position-if #'digit-char-p n))
                  n)))
      (format stream "#<MODE: ~@(~A~)>" name))))

(defmethod transpose ((object mode) offset 
                      &optional (scale *standard-scale*))
  ;(declare (ignore scale))
  (setf (slot-value object 'on) (scale-degree offset scale))
  object)

;;;
;;; force raw scale degree to modal member.
;;;

(defmethod force-mode (degree (mode mode))
  ;; force degree to mode by finding nearest mode member to it.
  (declare (fixnum degree) (optimize (speed 3) (safety 0)))
  (let ((i (mod (- degree
                   (or (slot-value mode 'on)
                       (error "~S is untransposed." mode)))
                (slot-value mode 'octave))))
    (declare (fixnum i))
    (+ degree (aref (slot-value mode 'fit) i))))

(defmethod mode-note (ref (mode mode))
  (let ((scale (slot-value mode 'scale)))
    (scale-note (force-mode (scale-degree ref scale) mode) scale)))

(defmethod mode-degree (ref (obj mode))
  ;(force-mode ref (slot-value obj 'scale))
  (force-mode ref obj) )

(defmethod mode-pitch (ref (mode mode))
  (let ((scale (slot-value mode 'scale)))
    (scale-pitch (force-mode (scale-degree ref scale) mode) scale)))

;;;
;;; return scale degree given legal modal degree
;;;

(defmethod unmodeify (degree (mode mode))
  ;; map mode degree to scale's interval system
  ;; i think this loses if mode's octave != scale's octave
  (let ((steps (slot-value mode 'steps))     ; mode octave
        (octave (slot-value mode 'octave)))  ; scale ocatave
    (multiple-value-bind (o i) (floor degree steps)
      (+ (* o octave) 
         (aref (slot-value mode 'int) i)
         (mod (or (slot-value mode 'on)
                  (error "~S is untransposed." mode))
              octave)))))

(defmethod scale-degree ((ref t) (mode mode))
  (scale-degree ref (slot-value mode 'scale)))

(defmethod scale-degree ((ref integer) (mode mode))
  (unmodeify ref mode))

(defmethod scale-note ((ref integer) (mode mode))
  (scale-note (unmodeify ref mode) (slot-value mode 'scale)))

(defmethod scale-pitch ((ref integer) (mode mode))
  (scale-pitch (unmodeify ref mode) (slot-value mode 'scale)))

;;;
;;; notes and pitches are filterd through.
;;;

(defmethod scale-note (ref (mode mode))
  ;(scale-note ref (slot-value mode 'scale))
  (mode-note ref mode))

(defmethod scale-pitch (ref (mode mode))
  ;(scale-pitch ref (slot-value mode 'scale))
  (mode-pitch ref mode))

;;;
;;; return modal degree given legal scaler degree. scaler degree
;;; must be a valid member in mode. Use force-mode to slam
;;; a scale degree to the nearest modal degree.
;;;

(defmethod modeify (degree (mode mode))
  ;; convert scale degree to mode's interval system.
  ;; i think this loses if mode Octave != scale Octave
  (let ((steps (slot-value mode 'steps))     ; mode octave
        (octave (slot-value mode 'octave))   ; scale octave
        (fit (slot-value mode 'fit)))
    (multiple-value-bind 
      (o i) 
      (floor (- degree 
                (mod (or (slot-value mode 'on)
                         (error "~S is untransposed." mode)) 
                     octave))
             octave)
      (unless (= 0 (aref fit i))
        (error "Scale step ~D is not a member of ~S." degree mode))
      (+ (* o steps)  ; if octaves arent = this fails
         i
         (loop for j from (1- i) downto 1 ; i and 0 are in mode
               sum (aref fit j))))))

#|
(defparameter m (defmode major 2 2 1 2 2 2 1))
(transpose m 'c4)
(setf x (notes 35 36 37 38 39 40 41 42 of m))
(read-items x)
(transpose m 'd4)
(read-items x)        ; degrees are relative to "key" of mode.

(transpose m 'c4)
(setf x (notes c4 ef f g af of m))  ; slam to nearest mode note
(read-items x)   

(transpose m 'c4)
(setf x (intervals 0 1 2 3 4 5 6 7 from 'fs4 of m returning note))
(read-items x)

(transpose m 'd4)
(read-items x)

(setf x (intervals 0 1 2 3 4 5 6 7
                   from (notes c4 fs4)
                   of m returning note))
(read-items x)
(read-items x)

(setf x (steps 1 for 8 from (notes c4 fs4)
               of m returning note))
(read-items x)
(read-items x)

(transpose m 'c4)
(loop for i from 35 to 50 collect (scale-note i m))
(loop for i from 35 to 50 collect (mode-note i m)) ;force scale to mode
(transpose m 'd4)
(loop for i from 35 to 50 collect (scale-note i m))
(loop for i from 35 to 50 collect (mode-note i m)) ;force scale to mode

(transpose m 'c4)
(loop repeat 20 collect (mode-note (between 60 79) m))
(transpose m 'fs4)
(loop repeat 20 collect (mode-note (between 60 79) m))
(transpose m 'b3)
(loop repeat 20 collect (mode-note (between 60 79) m))

(defmode major       2 2 1 2 2 2 1)
(defmode minor       2 1 2 2 1 2 2)
(defmode pentatonic  2 3 2 2 3)
(defmode whole-tone  2 2 2 2 2 2)
(defmode octatonic1  1 2 1 2 1 2 1 2)
(defmode octatonic2  2 1 2 1 2 1 2 1)
(defmode dorian      2 1 2 2 2 1 2)
(defmode phrygian    1 2 2 2 1 2 2)
(defmode lydian      2 2 2 1 2 2 1)
(defmode mixolydian  2 2 1 2 2 1 2)
(defmode aeolian     2 1 2 2 1 2 2)
(defmode locrian     1 2 2 1 2 2 2)
(defmode ionian      2 2 1 2 2 2 1)

(loop for n in '(major minor pentatonic whole-tone
                 octatonic1 octatonic2  dorian phrygian lydian 
                 mixolydian aeolian locrian ionian)
      for m = (transpose (find-scale n) 'd4)
      do (format t "~%~S:	~S" m
                 (loop for i from 48 repeat 24 
                      collect (mode-note i m))))

|#
